# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Minimal Flask application example for development.

SPHINX-START

Run the ElasticSearch and Redis server.

Run example development server:

.. code-block:: console

   $ pip install -e .[all]
   $ cd examples
   $ ./app-setup.sh
   $ ./app-fixtures.sh

Run example development server:

.. code-block:: console

   $ FLASK_DEBUG=1 FLASK_APP=app.py flask run -p 5000

Try to perform some search queries:

.. code-block:: console

   $ curl http://localhost:5000/?q=body:test

To be able to uninstall the example app:

.. code-block:: console

   $ ./app-teardown.sh

SPHINX-END
"""

from __future__ import absolute_import, print_function

import os

from elasticsearch import VERSION as ES_VERSION
from elasticsearch_dsl.query import Bool, Q, QueryString
from flask import Flask, jsonify, request
from flask_menu import Menu
from flask_security import current_user
from invenio_accounts import InvenioAccounts
from invenio_accounts.views import blueprint
from invenio_db import InvenioDB

from invenio_search import InvenioSearch, RecordsSearch

# Create Flask application
app = Flask(__name__)
app.config.update(
    ACCOUNTS_USE_CELERY=False,
    CELERY_ALWAYS_EAGER=True,
    CELERY_CACHE_BACKEND="memory",
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_RESULT_BACKEND="cache",
    MAIL_SUPPRESS_SEND=True,
    SECRET_KEY="CHANGE_ME",
    SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
    SQLALCHEMY_DATABASE_URI=os.getenv('SQLALCHEMY_DATABASE_URI',
                                      'sqlite:///instance/test.db'),
)
Menu(app)
InvenioDB(app)
InvenioAccounts(app)

app.register_blueprint(blueprint)

search = InvenioSearch(app)
search.register_mappings('demo', 'data')


class ExampleSearch(RecordsSearch):
    """Example search class."""

    class Meta:
        """Configuration for ``RecordsSearch`` class."""

        index = 'demo'
        doc_types = ['example']
        fields = ('*', )
        facets = {}

    def __init__(self, **kwargs):
        """Initialize instance."""
        super(ExampleSearch, self).__init__(**kwargs)
        if not current_user.is_authenticated:
            self.query = self.query._proxied & Q(
                Bool(filter=[Q('term', public=1)])
            )


@app.route('/', methods=['GET', 'POST'])
def index():
    """Query Elasticsearch using Invenio query syntax."""
    page = request.values.get('page', 1, type=int)
    size = request.values.get('size', 2, type=int)
    search = ExampleSearch()[(page - 1) * size:page * size]
    if 'q' in request.values:
        search = search.query(QueryString(query=request.values.get('q')))

    search = search.sort(
        request.values.get('sort', 'title')
    )
    search = ExampleSearch.faceted_search(search=search)
    results = search.execute().to_dict()
    return jsonify({'hits': results.get('hits')})



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for information retrieval."""

from __future__ import absolute_import, print_function

import errno
import json
import os
import warnings

from elasticsearch import VERSION as ES_VERSION
from pkg_resources import iter_entry_points, resource_filename, \
    resource_isdir, resource_listdir
from werkzeug.utils import cached_property

from . import config
from .cli import index as index_cmd
from .proxies import current_search_client
from .utils import build_index_name


def _get_indices(tree_or_filename):
    for name, value in tree_or_filename.items():
        if isinstance(value, dict):
            for result in _get_indices(value):
                yield result
        else:
            yield name


class _SearchState(object):
    """Store connection to elastic client and registered indexes."""

    def __init__(self, app,
                 entry_point_group_mappings=None,
                 entry_point_group_templates=None,
                 **kwargs):
        """Initialize state.

        :param app: An instance of :class:`~flask.app.Flask`.
        :param entry_point_group_mappings:
            The entrypoint group name to load mappings.
        :param entry_point_group_templates:
            The entrypoint group name to load templates.
        """
        self.app = app
        self.mappings = {}
        self.aliases = {}
        self.number_of_indexes = 0
        self._client = kwargs.get('client')
        self.entry_point_group_templates = entry_point_group_templates

        if entry_point_group_mappings:
            self.load_entry_point_group_mappings(entry_point_group_mappings)

    @cached_property
    def templates(self):
        result = None
        if self.entry_point_group_templates:
            result = self.load_entry_point_group_templates(
                self.entry_point_group_templates)
        return {k: v for d in result for k, v in d.items()} \
            if result is not None else {}

    def register_mappings(self, alias, package_name):
        """Register mappings from a package under given alias.

        :param alias: The alias.
        :param package_name: The package name.
        """
        # For backwards compatibility, we also allow for ES2 mappings to be
        # placed at the root level of the specified package path, and not in
        # the `<package-path>/v2` directory.
        if ES_VERSION[0] == 2:
            try:
                resource_listdir(package_name, 'v2')
                package_name += '.v2'
            except (OSError, IOError) as ex:
                if getattr(ex, 'errno', 0) != errno.ENOENT:
                    raise
                warnings.warn(
                    "Having mappings in a path which doesn't specify the "
                    "Elasticsearch version is deprecated. Please move your "
                    "mappings to a subfolder named according to the "
                    "Elasticsearch version which your mappings are intended "
                    "for. (e.g. '{}/v2/{}')".format(
                        package_name, alias),
                    PendingDeprecationWarning)
        else:
            package_name = '{}.v{}'.format(package_name, ES_VERSION[0])

        def _walk_dir(aliases, *parts):
            root_name = build_index_name(*parts)
            resource_name = os.path.join(*parts)

            if root_name not in aliases:
                self.number_of_indexes += 1

            data = aliases.get(root_name, {})

            for filename in resource_listdir(package_name, resource_name):
                index_name = build_index_name(*(parts + (filename, )))
                file_path = os.path.join(resource_name, filename)

                if resource_isdir(package_name, file_path):
                    _walk_dir(data, *(parts + (filename, )))
                    continue

                ext = os.path.splitext(filename)[1]
                if ext not in {'.json', }:
                    continue

                assert index_name not in data, 'Duplicate index'
                data[index_name] = self.mappings[index_name] = \
                    resource_filename(
                        package_name, os.path.join(resource_name, filename))
                self.number_of_indexes += 1

            aliases[root_name] = data

        # Start the recursion here:
        _walk_dir(self.aliases, alias)

    def register_templates(self, directory):
        """Register templates from the provided directory.

        :param directory: The templates directory.
        """
        try:
            resource_listdir(directory, 'v{}'.format(ES_VERSION[0]))
            directory = '{}/v{}'.format(directory, ES_VERSION[0])
        except (OSError, IOError) as ex:
            if getattr(ex, 'errno', 0) == errno.ENOENT:
                raise OSError(
                    "Please move your templates to a subfolder named "
                    "according to the Elasticsearch version "
                    "which your templates are intended "
                    "for. (e.g. '{}.v{}')".format(directory,
                                                  ES_VERSION[0]))
        result = {}
        module_name, parts = directory.split('.')[0], directory.split('.')[1:]
        parts = tuple(parts)

        def _walk_dir(parts):
            resource_name = os.path.join(*parts)

            for filename in resource_listdir(module_name, resource_name):
                template_name = build_index_name(*(parts[1:] + (filename, )))
                file_path = os.path.join(resource_name, filename)

                if resource_isdir(module_name, file_path):
                    _walk_dir((parts + (filename, )))
                    continue

                ext = os.path.splitext(filename)[1]
                if ext not in {'.json', }:
                    continue

                result[template_name] = resource_filename(
                    module_name, os.path.join(resource_name, filename))

        # Start the recursion here:
        _walk_dir(parts)
        return result

    def load_entry_point_group_mappings(self, entry_point_group_mappings):
        """Load actions from an entry point group."""
        for ep in iter_entry_points(group=entry_point_group_mappings):
            self.register_mappings(ep.name, ep.module_name)

    def load_entry_point_group_templates(self, entry_point_group_templates):
        """Load actions from an entry point group."""
        result = []
        for ep in iter_entry_points(group=entry_point_group_templates):
            with self.app.app_context():
                for template_dir in ep.load()():
                    result.append(self.register_templates(template_dir))
        return result

    def _client_builder(self):
        """Build Elasticsearch client."""
        from elasticsearch import Elasticsearch
        from elasticsearch.connection import RequestsHttpConnection

        return Elasticsearch(
            hosts=self.app.config.get('SEARCH_ELASTIC_HOSTS'),
            connection_class=RequestsHttpConnection,
        )

    @property
    def client(self):
        """Return client for current application."""
        if self._client is None:
            self._client = self._client_builder()
        return self._client

    def flush_and_refresh(self, index):
        """Flush and refresh one or more indices.

        .. warning::

           Do not call this method unless you know what you are doing. This
           method is only intended to be called during tests.
        """
        self.client.indices.flush(wait_if_ongoing=True, index=index)
        self.client.indices.refresh(index=index)
        self.client.cluster.health(
            wait_for_status='yellow', request_timeout=30)
        return True

    @property
    def cluster_version(self):
        """Get version of Elasticsearch running on the cluster."""
        versionstr = self.client.info()['version']['number']
        return [int(x) for x in versionstr.split('.')]

    @property
    def active_aliases(self):
        """Get a filtered list of aliases based on configuration.

        Returns aliases and their mappings that are defined in the
        `SEARCH_MAPPINGS` config variable. If the `SEARCH_MAPPINGS` is set to
        `None` (the default), all aliases are included.
        """
        whitelisted_aliases = self.app.config.get('SEARCH_MAPPINGS')
        if whitelisted_aliases is None:
            return self.aliases
        else:
            return {k: v for k, v in self.aliases.items()
                    if k in whitelisted_aliases}

    def create(self, ignore=None):
        """Yield tuple with created index name and responses from a client."""
        ignore = ignore or []

        def _create(tree_or_filename, alias=None):
            """Create indices and aliases by walking DFS."""
            # Iterate over aliases:
            for name, value in tree_or_filename.items():
                if isinstance(value, dict):
                    for result in _create(value, alias=name):
                        yield result
                else:
                    with open(value, 'r') as body:
                        yield name, self.client.indices.create(
                            index=name,
                            body=json.load(body),
                            ignore=ignore,
                        )

            if alias:
                yield alias, self.client.indices.put_alias(
                    index=list(_get_indices(tree_or_filename)),
                    name=alias,
                    ignore=ignore,
                )

        for result in _create(self.active_aliases):
            yield result

    def put_templates(self, ignore=None):
        """Yield tuple with registered template and response from client."""
        ignore = ignore or []

        def _put_template(template):
            """Put template in search client."""
            with open(self.templates[template], 'r') as body:
                return self.templates[template],\
                    current_search_client.indices.put_template(
                        name=template,
                        body=json.load(body),
                        ignore=ignore,
                )

        for template in self.templates:
            yield _put_template(template)

    def delete(self, ignore=None):
        """Yield tuple with deleted index name and responses from a client."""
        ignore = ignore or []

        def _delete(tree_or_filename, alias=None):
            """Delete indexes and aliases by walking DFS."""
            if alias:
                yield alias, self.client.indices.delete_alias(
                    index=list(_get_indices(tree_or_filename)),
                    name=alias,
                    ignore=ignore,
                )

            # Iterate over aliases:
            for name, value in tree_or_filename.items():
                if isinstance(value, dict):
                    for result in _delete(value, alias=name):
                        yield result
                else:
                    yield name, self.client.indices.delete(
                        index=name,
                        ignore=ignore,
                    )

        for result in _delete(self.active_aliases):
            yield result


class InvenioSearch(object):
    """Invenio-Search extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self._clients = {}

        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app,
                 entry_point_group_mappings='invenio_search.mappings',
                 entry_point_group_templates='invenio_search.templates',
                 **kwargs):
        """Flask application initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self.init_config(app)

        app.cli.add_command(index_cmd)

        state = _SearchState(
            app,
            entry_point_group_mappings=entry_point_group_mappings,
            entry_point_group_templates=entry_point_group_templates,
            **kwargs
        )
        self._state = app.extensions['invenio-search'] = state

    @staticmethod
    def init_config(app):
        """Initialize configuration.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        for k in dir(config):
            if k.startswith('SEARCH_'):
                app.config.setdefault(k, getattr(config, k))

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import hashlib

from elasticsearch_dsl import Q, Search
from flask import request

from invenio_search.api import DefaultFilter, RecordsSearch


def test_empty_query(app):
    """Test building an empty query."""
    with app.app_context():
        q = RecordsSearch()
        assert q.to_dict()['query'] == {'match_all': {}}

        q = RecordsSearch.faceted_search('')
        assert q._s.to_dict()['query'] == {'match_all': {}}

        q = RecordsSearch()[10]
        assert q.to_dict()['from'] == 10
        assert q.to_dict()['size'] == 1

        q = q[10:20]
        assert q.to_dict()['from'] == 10
        assert q.to_dict()['size'] == 10

        q = q.sort({'field1': {'order': 'asc'}})
        assert q.to_dict()['sort'][0] == {'field1': {'order': 'asc'}}

        q = q.sort()
        assert 'sort' not in q.to_dict()

        q = q.sort('-field1')
        assert q.to_dict()['sort'][0] == {'field1': {'order': 'desc'}}

        q = q.sort('field2', {'field3': {'order': 'asc'}})
        assert q.to_dict()['sort'][0] == 'field2'
        assert q.to_dict()['sort'][1] == {'field3': {'order': 'asc'}}
        q.sort()

        q = RecordsSearch()
        q = q.highlight('field1', index_options='offsets')
        assert len(q.to_dict()['highlight']['fields']) == 1
        assert q.to_dict()['highlight']['fields']['field1'] == {
            'index_options': 'offsets'
        }

        q = q.highlight('field2')
        assert len(q.to_dict()['highlight']['fields']) == 2
        assert q.to_dict()['highlight']['fields']['field1'] == {
            'index_options': 'offsets'
        }
        assert q.to_dict()['highlight']['fields']['field2'] == {}

        q = q.highlight()
        assert 'highligth' not in q.to_dict()


def test_elasticsearch_query(app):
    """Test building an empty query."""
    from flask import g

    class TestSearch(RecordsSearch):
        class Meta:
            default_filter = DefaultFilter(
                lambda: Q('terms', public=g.public)
            )

    with app.app_context():
        g.public = 1
        q = TestSearch()
        assert q.to_dict()['query'] == {
            'bool': {'minimum_should_match': "0<1",
                     'filter': [{'terms': {'public': 1}}]}
        }
        g.public = 0
        q = TestSearch()
        q = q.query(Q('match', title='Higgs'))
        assert q.to_dict()['query']['bool']['filter'] == [
            {'terms': {'public': 0}}
        ]
        assert q.to_dict()['query']['bool']['must'] == [
            {'match': {'title': 'Higgs'}}
        ]


class SpySearch(Search):
    exposed_params = {}

    def params(self, **kwargs):
        self.exposed_params.update(kwargs)
        return super(SpySearch, self).params(**kwargs)


def test_es_preference_param_no_request():
    """Test that the preference param is not added when not in a request."""
    RecordsSearch.__bases__ = (SpySearch,)

    rs = RecordsSearch()
    new_rs = rs.with_preference_param()
    assert new_rs.exposed_params == {}


def test_es_preference_param(app):
    """Test the preference param is correctly added in a request."""
    RecordsSearch.__bases__ = (SpySearch,)

    with app.test_request_context('/', headers={'User-Agent': 'Chrome'},
                                  environ_base={'REMOTE_ADDR': '212.54.1.8'}):
        rs = RecordsSearch()
        new_rs = rs.with_preference_param()

        alg = hashlib.md5()
        encoded_user_agent = 'Chrome'.encode('utf8')
        encoded_user_string = '{ip}-{ua}'.format(ip=request.remote_addr,
                                                 ua=encoded_user_agent)
        alg.update(encoded_user_string.encode('utf8'))
        digest = alg.hexdigest()

        assert new_rs.exposed_params == dict(preference=digest)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Configuration options for Invenio-Search."""

#
# ELASTIC configuration
#

SEARCH_ELASTIC_HOSTS = None  # default localhost
"""Elasticsearch hosts for the client.

By default the client connects to ``localhost:9200``. Below is an example of
connecting to three hosts via HTTPS with Basic authentication and switching of
SSL certificate verification:

.. code-block:: python

    params = dict(
        port=443,
        http_auth=('myuser', 'mypassword'),
        use_ssl=True,
        verify_certs=False,
    )
    SEARCH_ELASTIC_HOSTS = [
        dict(host='node1', **params),
        dict(host='node2', **params),
        dict(host='node3', **params),
    ]


The underlying library handles connection pooling and load-balancing between
the different nodes. Please see
`Elasticsearch client <https://elasticsearch-py.readthedocs.io/>`_
for further details.
"""

SEARCH_MAPPINGS = None  # loads all mappings and creates aliases for them
"""List of aliases for which, their search mappings should be created.

- If `None` all aliases (and their search mappings) defined through the
  ``invenio_search.mappings`` entry point in setup.py will be created.
- Provide an empty list ``[]`` if no aliases (or their search mappings)
  should be created.

For example if you don't want to create aliases
and their mappings for `authors`:

.. code-block:: python

    # in your `setup.py` you would specify:
    entry_points={
        'invenio_search.mappings': [
            'records = invenio_foo_bar.mappings',
            'authors = invenio_foo_bar.mappings',
        ],
    }

    # and in your config.py
    SEARCH_MAPPINGS = ['records']
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

r"""Elasticsearch management for Invenio.

Allows retrieving records from a configurable backend (currently, from
Elasticsearch).

Initialization
--------------

To be able to retrieve information from *somewhere*, we first need to setup
this *somewhere*. So make sure you have the correct version of Elasticsearch
installed and running (currently, version 2.x and 5.x are supported):

.. code-block:: console

   $ elasticsearch --version
   Version: 2.4.6, Build: 5376dca/2017-07-18T12:17:44Z, JVM: 1.8.0_144

In this case, we are using version 2 of Elasticsearch, so make sure to install
``invenio-search`` with the appropriate extras:

.. code-block:: console

    pip install invenio-search[elasticsearch2]

.. _creating_index:

Creating index
~~~~~~~~~~~~~~

To be able to run flask CLI commands later, we first need to have a Flask app,
so create the following ``app.py`` file:

.. code-block:: python

    # app.py
    from flask import Flask
    from invenio_search import InvenioSearch

    app = Flask('myapp')
    search = InvenioSearch(app)

This will create an empty index, which is not very useful, so let's add some
mappings. To not reinvent the wheel, let's reuse the mappings from the example
application. Add the following line at the end of ``app.py`` file that you
have just created:

.. code-block:: python

    # app.py
    ...
    search.register_mappings('demo', 'examples.data')

The above code will look into directory ``examples/data`` and load all the
mapping files it can find there into Elasticsearch (each file will create a new
index). It will also add ``demo`` as an alias to each of the indexes.

You can read more about the Elasticsearch mappings on elasticsearch website
(https://www.elastic.co/guide/en/elasticsearch/guide/current/mapping-intro.html).
Now we can finally create the index.

.. code-block:: console

   $ export FLASK_APP=app.py
   $ flask index init

If you are running Elasticsearch on your local machine, you can see that the
indices were created correctly by going to this URL:
http://localhost:9200/demo

Let's put some data inside

.. code-block:: console

   $ echo '{"title": "Hello invenio-search", "body": "test 1"}' | \
         flask index put demo-default-v1.0.0 example

No error message? Good! You can see that your new document was indexed by
going to this URL: http://localhost:9200/demo/_search

Searching for data
~~~~~~~~~~~~~~~~~~

Let's try to retrieve data in a programmatic way. Start a Python REPL and run
the following commands.

First, let's initialize app from the ``app.py`` file that we created in the
previous step.
Python REPL.

.. code-block:: python

   from app import app

Create a custom search class that will search for ``example`` type of
documents inside the ``demo`` index (or you could use the default
``RecordsSearch`` class to search for all document types in all indexes):

.. code-block:: python

   from invenio_search import RecordsSearch
   class ExampleSearch(RecordsSearch):
       class Meta:
           index = 'demo'
           doc_types = ['example']
           fields = ('*', )
           facets = {}
   search = ExampleSearch()

We will need a Flask application context, so let's push one:

.. code-block:: python

   app.app_context().push()

Let's find all documents:

.. code-block:: python

   response = search.execute()
   response.to_dict()

If everything went well, you should now see that we have 1 hit - a document
with ``Hello invenio-search`` title. If you get the ``TransportError(404,
'index_not_found_exception', 'no such index')`` error - it means that you
forgot to create the index (follow the steps from :ref:`creating_index` to
see how to setup an index and add example data).

Creating a search page
~~~~~~~~~~~~~~~~~~~~~~

Let's create a simple web page where you can send queries to the Elasticsearch
and see the results.
Create a new app.py file with a route.

.. code-block:: python

    # app.py
    from elasticsearch_dsl.query import QueryString
    from flask import Flask, jsonify, request
    from invenio_search import InvenioSearch, RecordsSearch

    app = Flask('myapp')

    search = InvenioSearch(app)

    # This line is needed to be able to call `flask index init`
    search.register_mappings('demo', 'examples.data')


    @app.route('/', methods=['GET', 'POST'])
    def index():
        search = RecordsSearch()
        if 'q' in request.values:
            search = search.query(QueryString(query=request.values.get('q')))

        return jsonify(search.execute().to_dict())

Run example development server:

.. code-block:: console

   $ FLASK_DEBUG=1 FLASK_APP=app.py flask run -p 5000

And now you can perform search queries:

.. code-block:: console

   $ curl http://localhost:5000/?q=body:test

Filtering
~~~~~~~~~
To filter out some documents, you can create your own search class. Let's try
to remove all private documents from the search results (by ``private``
documents, we understand all the documents that have ``public`` attribute set
to 0).

First, let's add one public and one private document to Elasticsearch:

.. code-block:: console

    $ echo '{"title": "Public", "body": "test 1", "public": 1}' | \
        flask index put demo-default-v1.0.0 example
    $ echo '{"title": "Private", "body": "test 2", "public": 0}' | \
        flask index put demo-default-v1.0.0 example

Now, create a new search class that will return all documents of type
``example`` from the ``demo`` index and select only the public ones
(documents where ``public`` is set to 1):

.. code-block:: python

    # app.py
    from elasticsearch_dsl.query import Bool, Q, QueryString

    class PublicSearch(RecordsSearch):
        class Meta:
            index = 'demo'
            doc_types = ['example']
            fields = ('*', )
            facets = {}

        def __init__(self, **kwargs):
            super(PublicSearch, self).__init__(**kwargs)
            self.query = self.query._proxied & Q(
                Bool(filter=[Q('term', public=1)])
            )

Update the ``index`` function and replace the search class with our new
``PublicSearch`` class:

.. code-block:: python

    # app.py
    @app.route('/', methods=['GET', 'POST'])
    def index():
        search = PublicSearch()
        ...

Now, you can search for documents with ``test`` in the body.

.. code-block:: console

   $ curl http://localhost:5000/?q=body:test

You should find only one document - the one with ``Public`` title.

This is a very simple example of how to filter out some records. If you want
to see how to hide some records if the user is not logged in, check out the
:ref:`examplesapp`. If you want to define role based access rights control,
check the invenio-access_ module.


Miscellaneous
-------------

Elasticsearch version support
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Because of the breaking changes that are introduced in Elasticsearch between
major versions in relation to mappings, a specific directory structure has to
be followed, in order to specify which JSON mapping files will be used for
creating the Elasticsearch indices. For backwards compatibility with existing
Invenio modules and installations, for Elasticsearch 2, mappings will be loaded
from the root level of the package directory. You can see a full example in the
``examples/data`` directory of the ``invenio-search`` repository:

.. code-block:: console

    $ tree --dirsfirst examples/data

    examples/data
    +- demo            # Elasticsearch 2 mappings
    |  +- authorities
    |  |  +- authority-v1.0.0.json
    |  +- bibliographic
    |  |  +- bibliographic-v1.0.0.json
    |  +- default-v1.0.0.json
    +- v5
    |  +- demo        # Elasticsearch 5 mappings
    |  |  +- authorities
    |  |  |  +- authority-v1.0.0.json
    |  |  +- bibliographic
    |  |  |  +- bibliographic-v1.0.0.json
    |  |  +- default-v1.0.0.json
    |  +- __init__.py
    +-- __init__.py


Elasticsearch plugins
~~~~~~~~~~~~~~~~~~~~~

For convenience, you can install a plugin like Elastic HQ
(http://www.elastichq.org/) for easy introspection of your indexes and their
content. Otherwise, you can use curl (as described here:
https://www.elastic.co/guide/en/elasticsearch/guide/current/_talking_to_elasticsearch.html).

.. _invenio-access:  https://invenio-access.readthedocs.io/

"""

from __future__ import absolute_import, print_function

from .api import RecordsSearch
from .ext import InvenioSearch
from .proxies import current_search, current_search_client
from .version import __version__

__all__ = (
    '__version__',
    'InvenioSearch',
    'RecordsSearch',
    'current_search',
    'current_search_client',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Test CLI."""

from __future__ import absolute_import, print_function

import ast

from click.testing import CliRunner
from elasticsearch import VERSION as ES_VERSION
from flask.cli import ScriptInfo
from mock import patch

from invenio_search.cli import index as cmd
from invenio_search.proxies import current_search_client


def test_init(app, template_entrypoints):
    """Run client initialization."""
    search = app.extensions['invenio-search']
    search.register_mappings('records', 'mock_module.mappings')

    assert 'records' in search.aliases
    assert set(search.aliases['records']) == set([
        'records-authorities',
        'records-bibliographic',
        'records-default-v1.0.0',
    ])
    assert set(search.mappings.keys()) == set([
        'records-authorities-authority-v1.0.0',
        'records-bibliographic-bibliographic-v1.0.0',
        'records-default-v1.0.0',
    ])
    assert 6 == search.number_of_indexes

    with patch('invenio_search.ext.iter_entry_points',
               return_value=template_entrypoints('invenio_search.templates')):
        if ES_VERSION[0] == 2:
            assert len(search.templates.keys()) == 2
            assert 'record-view-v1' in search.templates
            assert 'subdirectory-file-download-v1' in search.templates
        else:
            assert len(search.templates.keys()) == 1
            assert 'record-view-v{}'.format(ES_VERSION[0]) in search.templates

    with app.app_context():

        current_search_client.indices.delete_alias('_all', '_all',
                                                   ignore=[400, 404])
        current_search_client.indices.delete('*')
        aliases = current_search_client.indices.get_alias()
        assert 0 == len(aliases)

    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)

    with runner.isolated_filesystem():
        result = runner.invoke(cmd, ['init', '--force'],
                               obj=script_info)
        with app.app_context():
            if ES_VERSION[0] == 2:
                assert current_search_client.indices.exists_template(
                    'subdirectory-file-download-v1')
                assert current_search_client.indices.exists_template(
                    'record-view-v1')
            else:
                assert current_search_client.indices.exists_template(
                    'record-view-v{}'.format(ES_VERSION[0]))
        assert 0 == result.exit_code

    with app.app_context():
        aliases = current_search_client.indices.get_alias()
        assert 5 == sum(len(idx.get('aliases', {}))
                        for idx in aliases.values())

        assert current_search_client.indices.exists(
            list(search.mappings.keys())
        )

    # Clean-up:
    with app.app_context():
        result = runner.invoke(cmd, ['destroy'],
                               obj=script_info)
        assert 1 == result.exit_code

        result = runner.invoke(cmd, ['destroy', '--yes-i-know'],
                               obj=script_info)
        assert 0 == result.exit_code

        aliases = current_search_client.indices.get_alias()
        assert 0 == len(aliases)


def test_list(app):
    """Run listing of mappings."""
    original = app.config.get('SEARCH_MAPPINGS')
    app.config['SEARCH_MAPPINGS'] = ['records']
    search = app.extensions['invenio-search']
    search.register_mappings('authors', 'mock_module.mappings')
    search.register_mappings('records', 'mock_module.mappings')

    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)

    result = runner.invoke(cmd, ['list', '--only-aliases'], obj=script_info)
    # Turn cli outputted str presentation of Python list into a list
    assert set(ast.literal_eval(result.output)) == set(['records', 'authors'])

    result = runner.invoke(cmd, ['list'], obj=script_info)
    assert result.output == (
        u"├──authors\n"
        u"│  └──authors-authors-v1.0.0\n"
        u"└──records *\n"
        u"   ├──records-authorities\n"
        u"   │  └──records-authorities-authority-v1.0.0\n"
        u"   ├──records-bibliographic\n"
        u"   │  └──records-bibliographic-bibliographic-v1.0.0\n"
        u"   └──records-default-v1.0.0\n\n"
    )


def test_check(app):
    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)

    result = runner.invoke(cmd, ['check'], obj=script_info)
    assert result.exit_code == 0

    with patch('invenio_search.cli.ES_VERSION',
               return_value=(ES_VERSION[0] + 1, 0, 0)):
        result = runner.invoke(cmd, ['check'], obj=script_info)
        assert result.exit_code != 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility functions for search engine."""

import os


def build_index_name(*parts):
    """Build an index name from parts.

    :param parts: Parts that should be combined to make an index name.
    """
    return os.path.splitext('-'.join([part for part in parts if part]))[0]


def schema_to_index(schema, index_names=None):
    """Get index/doc_type given a schema URL.

    >>> from invenio_search.utils import schema_to_index
    >>> schema_to_index('records/record-v1.0.0.json')
    ('records-record-v1.0.0', 'record-v1.0.0')
    >>> schema_to_index('default-v1.0.0.json')
    ('default-v1.0.0', 'default-v1.0.0')
    >>> schema_to_index('default-v1.0.0.json', index_names=[])
    (None, None)
    >>> schema_to_index('invalidextension')
    (None, None)

    :param schema: The schema name
    :param index_names: A list of index name.
    :returns: A tuple containing (index, doc_type).
    """
    parts = schema.split('/')
    doc_type = os.path.splitext(parts[-1])

    if doc_type[1] not in {'.json', }:
        return (None, None)

    if index_names is None:
        return (build_index_name(*parts), doc_type[0])

    for start in range(len(parts)):
        index_name = build_index_name(*parts[start:])
        if index_name in index_names:
            return (index_name, doc_type[0])

    return (None, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import pytest
from elasticsearch import VERSION as ES_VERSION
from flask import Flask
from mock import patch

from invenio_search import InvenioSearch, current_search, current_search_client
from invenio_search.utils import schema_to_index


def test_version():
    """Test version import."""
    from invenio_search import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = InvenioSearch(app)
    assert 'invenio-search' in app.extensions

    app = Flask('testapp')
    ext = InvenioSearch()
    assert 'invenio-search' not in app.extensions
    ext.init_app(app)
    assert 'invenio-search' in app.extensions


def test_flush_and_refresh(app):
    """Test flush and refresh."""
    search = app.extensions['invenio-search']
    search.flush_and_refresh('_all')


def test_client_reference():
    """Test client reference."""
    client1 = {'name': 'client1'}
    client2 = {'name': 'client2'}

    app1 = Flask('testapp1')
    app2 = Flask('testapp2')

    ext = InvenioSearch()
    assert 'invenio-search' not in app1.extensions
    assert 'invenio-search' not in app2.extensions

    ext.init_app(app1, client=client1)
    assert 'invenio-search' in app1.extensions

    ext.init_app(app2, client=client2)
    assert 'invenio-search' in app2.extensions

    with app1.app_context():
        assert current_search_client == client1

    with app2.app_context():
        assert current_search_client == client2


def test_default_client(app):
    """Test default client."""
    with app.app_context():
        current_search_client.cluster.health(
            wait_for_status='yellow', request_timeout=1
        )


@pytest.mark.parametrize(('schema_url', 'result'), [
    ('invalidfileextension',
     (None, None)),
    ('records/record-v1.0.0.json',
     ('records-record-v1.0.0', 'record-v1.0.0')),
    ('/records/record-v1.0.0.json',
     ('records-record-v1.0.0', 'record-v1.0.0')),
])
def test_schema_to_index(schema_url, result):
    """Test conversion of schema to index name and document type."""
    assert result == schema_to_index(schema_url)


def test_load_entry_point_group(template_entrypoints):
    """Test entry point loading."""
    app = Flask('testapp')
    ext = InvenioSearch(app)
    ep_group = 'test'

    def mock_entry_points_mappings(group=None):
        assert group == ep_group

        class ep(object):

            name = 'records'
            module_name = 'mock_module.mappings'

        yield ep

    assert len(ext.mappings) == 0
    with patch('invenio_search.ext.iter_entry_points',
               mock_entry_points_mappings):
        ext.load_entry_point_group_mappings(
            entry_point_group_mappings=ep_group)
    assert len(ext.mappings) == 3
    # Check that mappings are loaded from the correct folder depending on the
    # ES version.
    if ES_VERSION[0] > 2:
        mappings_dir = 'mock_module/mappings/v{}/records'.format(ES_VERSION[0])
        assert all(mappings_dir in path for path in ext.mappings.values())

    with patch('invenio_search.ext.iter_entry_points',
               return_value=template_entrypoints('invenio_search.templates')):
        if ES_VERSION[0] == 2:
            assert len(ext.templates.keys()) == 2
        elif ES_VERSION[0] == 5:
            assert len(ext.templates.keys()) == 1


@pytest.mark.parametrize(('aliases_config', 'expected_aliases'), [
    (None, ['records', 'authors']),
    (['records', 'authors'], ['records', 'authors']),
    (['authors'], ['authors']),
    ([], []),
    (['does_not_exist'], []),
])
def test_whitelisted_aliases(app, aliases_config, expected_aliases):
    """Test functionality of active aliases configuration variable."""

    orig = app.config['SEARCH_MAPPINGS']

    search = app.extensions['invenio-search']
    search.register_mappings('records', 'mock_module.mappings')
    search.register_mappings('authors', 'mock_module.mappings')

    app.config.update(SEARCH_MAPPINGS=aliases_config)

    with app.app_context():
        current_search_client.indices.delete_alias('_all', '_all',
                                                   ignore=[400, 404])
        current_search_client.indices.delete('*')
        list(current_search.create(ignore=None))

        aliases = current_search_client.indices.get_alias()
        if expected_aliases == []:
            assert 0 == len(aliases)
        else:
            assert current_search_client.indices.exists(expected_aliases)

    app.config['SEARCH_MAPPINGS'] = orig
